
# +++ Dokumentation fbdraw +++

Stand: Sa 29. Apr 20:28:51 CEST 2023

Sinn dieser Tools ist es, framebuffer Grafik im Userpace
einfacher verfügbar zu machen und den Zugriff
scriptbar zu gestalten.

# Compilieren

Einfach "make" eintippen.
Der Quellcode ist optimiert zur Nutzung mit musl-gcc.
In der Makefile wird daher auch CC=musl-gcc gesetzt.

# Beispiele

make
make test
make demo

# Dokumentation

## fbenv

Öffnet das framebuffer device, liest dessen
Auflösung und Farbtiefe aus und überträgt diese
Information in die nachfolgenden Umgebungsvariablen

"SCREEN_HEIGHT"
"SCREEN_WIDTH"
"SCREEN_COLORDEPTH"

Bestehende Werte werden dabei überschrieben.
Dann forked das Tool einen child prozess mit diesem
geänderten Environment. 

Beispielaufrufe: 

fbenv env | grep SCREEN
fbenv drawfb < demodata.txt
fbenv sh
W=$(fbenv sh -c 'echo "$SCREEN_WIDTH"')

## fbdraw

Zeichnet in den Framebuffer.

Das Tool liest zeilenweise Befehle und Parameter von stdin und
klont diese Zeile auch unverändert zu stdout. Es werden nur
die ersten 63 Zeichen einer Zeile interpretiert. Die restlichen
Zeichen pro Zeile werden verworfen.

Beispiele siehe demodata.txt

Folgende Befehle existieren:

cmd	argumente	name	description
-----------------------------------------------------------------------------------
c	r g b		color	Setzt die Zeichenfarbe nach R G B
m	x y		move	Bewegt den Zeichencursor zu gewünschten Koordinaten
r	dx dy		rect	Zeichnet ein Rechteck
p	x y r g b 	pixel	Zeichnet exakt einen Pixel an x y in Farbe r g b
s	n		sleep	Wartet angegebene Anzahl an Microsekunden
o	ox oy		offset	Definiert globalen Offset für Zeichenoperationen		

Befehle und Parameter/Argumente werden durch ein Leerzeichen getrennt.
Alle Argumente MÜSSEN Ganzzahlen (integer) sein.

Beispiele:

c 140 20 0		Wechselt zur Farbe rot
m 20 80			Wechselt zu den Koordinaten x=20 und y=80
r 10 10			Zeichnet ein Quadrat 10x10 px
r 10 40			Zeichnet ein Rechteck 10px breit und 40px hoch
s 50000			Wartet 50 Millisekunden (für Animationen)
p 23 42 164 163 255	Zeichnet einen hellblauen Pixel an x=23 y=42
o 50 100		Definiert einen globalen Offset von x=50 y=100
 
Beispielaufruf: fbdraw < demodata.txt > /dev/null

Kommentarzeilen können mit einem # beginnen und werden ignoriert.
Auch alle unbekannten Befehle werden ignoriert.

Die Parametrisierung von fbdraw erfolgt über *Umgebungsvariablen*
 
Als Framebuffer Device wird standardmäßig /dev/fb0 verwendet,
sofern die entsprechende Umgebungsvariable "FBDEV" nicht gesetzt ist.
Du kannst ein anderes Framebuffer Device verwenden, indem du

export FBDEV=/dev/fb1

setzt, bzw. auch direkt vor dem Aufruf von fbdraw mit angibst

FBDEV=/dev/fb1 fbdraw < demodata.txt > /dev/null

Auch die Bildschirmauflösung wird via Umgebungsvariablen so gesetzt:

export SCREEN_WIDTH=1920
export SCREEN_HEIGHT=1080
export SCREEN_COLORDEPTH=32

Durch die Klon-Funktion kannst du die Ausgabe auch auf mehrere Devices
zugleich zeichnen und unterschiedliche Auflösung verwenden


cat demodata.txt | \
FBDEV=/dev/fb0 fbenv fbdraw | \
FBDEV=/dev/fb1 SCREEN_COLORDEPTH=16 SCREEN_WIDTH=1024 SCREEN_HEIGHT=768 fbdraw | \
FBDEV=/dev/fb2 SCREEN_COLORDEPTH=32 SCREEN_WIDTH=800 SCREEN_HEIGHT=480 fbdraw | \
> /dev/null

fbdraw verwirft Zeichen-Anweisungen die außerhalb desjeweiligen
Bildschirmbereichs liegen.

## fbmouse

Gibt die Cursor-Koordinaten der Maus auf stdout aus.
Zudem werden die Events "Linksklick", "Rechtsklick", "Mittelklick"
und die Scrollrad-Bewegungen "hoch" und "runter" erfasst und als
Event ausgegeben. Jedem Event wird dabei das Zeichen "#" vorangestellt,
sodass er in Scripten sehr gut filterbar ist. Zudem wird bei events
auch ein 64 px Grid Raster berücksichtigt und die tile id genannt.
Siehe dazu die weiterführenden Beispiele.

Die Parametrisierung von fbmouse erfolgt über *Umgebungsvariablen*

export SCREEN_WIDTH=1920
export SCREEN_HEIGHT=1080
export MOUSEDEVICE=/dev/input/mice

Die Bildschirmgröße ist wichtig, um die korrekte Tile ID zu erzeugen
und auch die Maximalwerte zu definieren.

Du kannst auch hier fbenv verwenden, um die Umgebungsvariablen
"SCREEN_WIDTH" und "SCREEN_HEIGHT" automatisch zu setzen

MOUSEDEVICE=/dev/input/mice fbenv fbmouse

Events: click_left, click_right, click_middle, scroll_up, scroll_down

Ausgabeformat bei Klick/Scroll: "# event tile_id x y"
Ausgabeformat bei Bewegung der Maus: "x y"  (ohne führende Raute)


## fbtouch

Analog zu fbmouse, allerdings für TouchScreen Events, die
statt *relativen Koordinaten* stets *absolute Korrdinaten* liefern
und vollständig anders abgefragt werden.

Die Parametrisierung von fbtouch erfolgt über *Umgebungsvariablen*

export SCREEN_WIDTH=800
export SCREEN_HEIGHT=400
export TOUCHDEVICE=/dev/input/event0

Optional: export MULTITOUCH=1

Beachte, dass du das richtige Event Device wählen musst!
Du kannst dazu auch Hersteller und Modellname wählen.
Siehe:

ls -1 /dev/input/by-id/

Die Bildschirmgröße ist wichtig, um die korrekte Tile ID zu erzeugen
und auch die Maximalwerte zu definieren.

Du kannst auch hier fbenv verwenden, um die Umgebungsvariablen
"SCREEN_WIDTH" und "SCREEN_HEIGHT" automatisch zu setzen.
Ein Beispielaufruf von fbtouch sieht also wie folgt aus:

TOUCHDEVICE=/dev/input/event0 fbenv fbtouch

Bei einem Multitouch Device musst du diese Umgebungsvariable setzen

MULTITOUCH=1 TOUCHDEVICE=/dev/input/event0 fbenv fbtouch

Das Ausgabeformat des Tools auf stdout ist:  "# touch tile_id x y"

# Weiterführende Beispiele

fbenv fbmouse | grep "#" | cut -d " " -f 3

zeigt ledliglich Events an (z.B. Clicks oder Scrollrad-Ereignisse)
und nennt dabei jeweils die tile ID. die tile ID bezieht sich
auf deinen grid aus 64 Pixeln, also ein virtuelles festes Raster
von 64x64 Pixeln, welches über den Bildschirmbereich gelegt wird.

Beispiel bei einer Auflösung von 1920x1080
30 Blöcke in der Horizontalen, in dem durch 64 geteilt wird:
(1920/64)+1=30+1=31 .. wir beginnen bei ID 1 statt ID 0

```
[ 1][ 2][ 3][ 4][ 5]...[31]
[32][33]...............[60]
.                         .
......................[512]
```
Um also zu sehen, ob das Grid an dem tile mit der ID 128
mit linker Maustaste geklickt wurde, musst du nur
derartiges verwenden:

fbenv fbmouse | stdbuf -i0 -o0 grep -q "# click_left 128" && echo ok


fbenv fbtouch | stdbuf -i0 -o0 grep -q "# touch 1" && echo "klick in linke obere ecke erkannt"

# Tipps und Tricks

## Wie kann ich eigene Pixel-Grafiken in das fbdraw p draw Format konvertieren?

Nutze das Shell Script "convert2fbdraw.sh" im Unterordner "examples".
Es verwendet imagemagick, um aus bekannten Grafikformaten die rgb Werte
(links oben nach rechts unten) zu extrahieren, daraus dann zeilenweise
p Format Zeilen auf stdout zu erstellen. Das Script benötigt zudem
xxd und gawk.

Tipp: Um Speicherplatz zu sparen und auch den sp
teren Zeichenprozess zu beschleunigen, solltest
du auch mit transparenten Hintergründen arbeiten und nur die Pixel
zeichnen, die wirklich benötigt werden. Dazu wird einfach eine
gewünschte Farbe, z.B. schwarz "0 0 0" dann mit grep herausgefiltert.

cat beispiel.jpg | convert2fbdraw.sh 640 | grep -v "^p .* 0 0 0$" | gzip -9 -c > result.txt.gz


## Ich möchte eine Hintergrundgrafik schnell im Framebuffer zeichnen. Wie mache ich das?

Der zeilenweise Aufbau mit p Befehlen dafür recht langsam ist, solltest du
zunächst einmal das Bild damit komplett in den Framebuffer zeichnen.
Danach speicherst du den kompletten Framebuffer deines Geräts in eine
komprimierte Datei. Dann kannst du jederzeit den Framebuffer damit wieder 
initial überschreiben, bevor du weitere fbdraw Befehle wieder darauf anwendest.

Beispiel:

Nehmen wir an, du hast einen Hintergrund von 800 480 Pixel in 
der Datei background.jpg. Dann zeichnen wir diese in den Framebuffer
und speichern diese dann in background.fb.gz. Dann können wir jederzeit
wieder mit zcat den Framebuffer überschreiben. Dafür benötigen wir auch
kein fbdraw. Es ist allerdings dann fix für dieses Gerät, Farbtiefe und
Auflösung hinterlegt!!

convert2fbdraw.sh 800 < background.jpg | fbenv fbdraw

cat /dev/fb0 | gzip -9 -c > background_800x480.fb.gz
zcat background_800x480.fb.gz > /dev/fb0


